library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use work.packet.all;

-- interface do ArbitroParaleloRotativo
entity ArbitroParaleloRotativo is
	port
	(
		clock:		in std_logic;
		reset:		in std_logic;
		ack_rot:	in std_logic;
		h:			in std_logic_vector(4 downto 0);
		ack_h:		out std_logic_vector(4 downto 0);
		incoming:	out std_logic_vector(2 downto 0);
		req_rot:	out std_logic
	);
end ArbitroParaleloRotativo;

-- implementao do ArbitroParaleloRotativo
architecture ArbitroParaleloRotativo of ArbitroParaleloRotativo is

type state is (S0,S1,S2,S3,S4);
signal ES, PES: state;
signal ask: std_logic;
signal counter,timeout: std_logic_vector(5 downto 0);
signal sel,prox: integer; 
begin

	timeout <= "110010"; -- timeout igual a 50 ciclos ou 1000ns.
	
	ask <= '1' when h(LOCAL)='1' or h(EAST)='1' or h(WEST)='1' or h(NORTH)='1' or h(SOUTH)='1' else '0';
	
		
	process(reset,clock)
	begin
		if reset='1' then
			ES<=S0;
		elsif clock'event and clock='1' then
			ES<=PES;
		end if;
	end process;

	process(sel,h)
	begin
		case sel is
			when LOCAL=> 
				if h(EAST)='1' then prox<=EAST;
				elsif h(WEST)='1' then	prox<=WEST;
				elsif h(NORTH)='1' then	prox<=NORTH;
				elsif h(SOUTH)='1' then	prox<=SOUTH;
				elsif h(LOCAL)='1' then	prox<=LOCAL;
				else prox<=EAST; end if;
			when EAST=> 
				if h(WEST)='1' then	prox<=WEST;
				elsif h(NORTH)='1' then prox<=NORTH;
				elsif h(SOUTH)='1' then prox<=SOUTH;
				elsif h(LOCAL)='1' then prox<=LOCAL;
				elsif h(EAST)='1' then prox<=EAST;
				else prox<=WEST; end if;
			when WEST=> 
				if h(NORTH)='1' then prox<=NORTH;
				elsif h(SOUTH)='1' then prox<=SOUTH;
				elsif h(LOCAL)='1' then prox<=LOCAL;
				elsif h(EAST)='1' then prox<=EAST;
				elsif h(WEST)='1' then prox<=WEST;
				else prox<=NORTH; end if;
			when NORTH=> 
				if h(SOUTH)='1' then prox<=SOUTH;
				elsif h(LOCAL)='1' then prox<=LOCAL;
				elsif h(EAST)='1' then prox<=EAST;
				elsif h(WEST)='1' then prox<=WEST;
				elsif h(NORTH)='1' then prox<=NORTH;
				else prox<=SOUTH; end if;
			when SOUTH=> 
				if h(LOCAL)='1' then prox<=LOCAL;
				elsif h(EAST)='1' then prox<=EAST;
				elsif h(WEST)='1' then prox<=WEST;
				elsif h(NORTH)='1' then prox<=NORTH;
				else prox<=SOUTH; end if;
			when others => prox<=LOCAL;
		end case;
	end process;		
			
	------------------------------------------------------------------------------------------------------	
	-- PARTE COMBINACIONAL PARA DEFINIR o prximo estado da mquina.
	------------------------------------------------------------------------------------------------------	
	--S0 -> Zera variveis
	--S1 -> Arbitro verifica se alguem requisitou roteamento
	--S2 -> Requisita roteamento, envia porta de entrada e endereco destino
	--S3 -> Espera resposta de roteamento OK ou cai fora quando expirar timeout
	--S4 -> Envia ack confirmando roteamento
	process(ES,ask,sel,counter,timeout,ack_rot) --sel nao precisa
	begin
		case ES is
			when S0 => PES <= S1;
			when S1 => if ask='1' then PES <= S2; else PES <= S1; end if;
			when S2 => PES <= S3;
			when S3 => if ack_rot='1' then PES <= S4;
				elsif counter=timeout then PES<= S1;
				else PES <= S3;
				end if;
			when S4 => PES <= S1;
		end case;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- executa as aes correspondente ao estado atual da mquina de estados
	------------------------------------------------------------------------------------------------------	
	process (clock)
	begin
		if clock'event and clock='0' then
			case ES is
				-- Zera variveis
				when S0 =>
					sel <= 0;
					incoming <= (others => '0');
					req_rot <= '0';
					ack_h <= (others => '0');
					counter <= (others => '0');
				-- Chegou um header
				when S1=>
					ack_h <= (others => '0');
					counter <= (others => '0');
				-- Seleciona quem tera direito a requisitar roteamento
				when S2=>
					sel <= prox;
				-- Espera resposta de roteamento OK ou cai fora quando expirar timeout
				when S3=>
					-- Requisita roteamento
					req_rot <= '1';
					incoming <= conv_vector(sel);
					counter <= counter + '1';
				-- Envia ack confirmando roteamento
				when S4=>
					req_rot <= '0';
					ack_h(sel) <= '1';
				end case;
		end if;
	end process;	

end ArbitroParaleloRotativo;
