library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use work.packet.all;

-- interface da Fila
entity Fila is
        port(
        clock:          in std_logic;
        reset:          in std_logic;
        data_in:        in reg8;
        rx:             in std_logic;
        ack_rx:         out std_logic;
        h:              out std_logic;
        ack_h:          in std_logic;
        data_av:        out std_logic;
        data:           out reg8;
        data_ack:       in std_logic;
        free:           out std_logic);
end Fila;

architecture Fila of Fila is

 type fila_out is (S1, S2, S_SIZE, S_HEADER, S_END);
 signal EA : fila_out;

 signal buf: buff;
 signal first,last: reg3;

 signal tem_espaco_na_fila : std_logic;
 signal counter_flit : reg8;
begin 
        
   -- entrada de dados na fila. Sempre que ver rx e tiver lugar na fila armazena o dado
		 
   --tem_espaco_na_fila <= '1' when not((first=x"0" and last=x"7") or (first=last+1)) else '0';

   process(reset, clock)
     begin
         if reset='1' then
                tem_espaco_na_fila <= '1';  
         elsif clock'event and clock='0' then
         	if not((first=x"0" and last=x"7") or (first=last+1)) then
                  tem_espaco_na_fila <= '1'; 
              else 
                  tem_espaco_na_fila <= '0';
   		end if;
         end if;
   end process;
   
   ack_rx <= '1' when  (tem_espaco_na_fila='1' and rx='1') else '0';
                  
   process(reset, clock)
     begin
         if reset='1' then
                last <= (others=>'0');  
         elsif clock'event and clock='1' then
              if tem_espaco_na_fila='1' and rx='1' then
                 buf(CONV_INTEGER(last)) <= data_in;
                 last <= last + 1;
              end if;
         end if;
   end process;
   
   
   --  tira da fila QUANDO se detecta presena de dados na fila => DESCOMPATIBILIZA
   --  ENTRADA DE SADA = muito importante!!
   -- 
   --  para sabermos se tem dado na fila basta ver se top!=botton
   -- 
   --  falta dois mecanismos fceis de implementar: time_out e flush, j tenho uma idia
   --  simples de implemntar
  
  --- o valor lido da fila  funo do ponteiro first
  data <= buf(CONV_INTEGER(first));
  
  process(reset, clock)
    begin
      if reset='1' then
            data_av <= '0';
            first <= (others=>'0');
            EA <= S1;
            h <= '0';
            counter_flit <= (others=>'0');
      elsif clock'event and clock='1' then
      
         case EA is                  
           when S1 =>  counter_flit <= "00000000";
                       h<='0';                 
                       data_av <= '0';
                       if first /= last then        -- detectou dado na fila
                             EA <= S_HEADER;
                       else
                             EA<=S1;
                       end if;
  
  
           when S_HEADER =>  if ack_h='1' then
                                 EA <= S_SIZE ;      -- depois de rotear envia o pacote 
                                 h<='0';
                             else
                                 EA <= S_HEADER;
                                 h<='1';             -- pede roteamento
                             end if;                 -- FALTA_TIME OUT E FLUSH
  
           when S_SIZE  => if data_ack='1' then        -- mantm este dado at a resposta
                                first <= first+1;      -- assim que vier a resposta tira do buffer
                                EA <= S2;
                                data_av <= '0';   
                           else
                                EA <= S_SIZE;
                                data_av <= '1';
                           end if;
  
           when S2 => if data_ack='1' then                 -- mantm este dado at a resposta
                            first <= first+1;             -- assim que vier a resposta tira do buffer
                            EA <= S2;
                            data_av <= '0'; 
                            
                            if counter_flit = "00000000" then
                                    counter_flit <=  buf(CONV_INTEGER(first)); 
                            else
                                    counter_flit <= counter_flit - 1; 
                            end if;
                                          
                      elsif counter_flit = "00000001" then
                            EA <= S_END;
                            data_av <= '1';
                                       
                      else
                            EA <= S2;
                            data_av <= '1';
                      end if;
                  
           when S_END => if data_ack='1' then                 -- mantm este dado at a resposta
                            first <= first+1;             -- assim que vier a resposta tira do buffer
                            EA <= S1;
                            data_av <= '0'; 
                         else
                            EA <= S_END;
                            data_av <= '1';
                        end if;
                   
           
                       --   EA <= S1;
                       --  data_av <= '0';  
                       --  first <= first+1;
                         
         end case;
      end if;
  end process;            


  --
  -- CONTROLE PARA INDICAR DADOS EM RECEPO NO BUFFER - free=1 livre  - free=0 ocpupado
  --
  process(reset, clock)
    begin
      if reset='1' then
            free <=  '1';
       elsif clock'event and clock='1' then
          
           if ack_h = '1' then
                  free <= '0';
           --elsif EA=S_END then   
		   elsif EA=S1 then
                  free <= '1';
           end if;
       end if;
  end process;

  
end Fila;