---------------------------------------------------------------------------------------	
--                               SENDRECEIVE PARA NOC
---------------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use work.packet.all;

-- interface do Send e do Receive
entity SendReceive is
	port(
	-- debug
	state:		out std_logic_vector(141 downto 0);

	clock: 		in std_logic;
	reset: 		in std_logic;
	addressC:	in reg8;
	avSend:		in std_logic;
	nwordSend:	in reg3;
	busySend:	out std_logic;
	busyWrapper:in std_logic;
	dataSend:	in packet;
	dataReceive:out packet;
	avReceive:	out std_logic;
	rx:			in std_logic;
	ack_rx:		out std_logic;
	data_in:	in reg8;
	tx:			out std_logic;
	ack_tx:		in std_logic;
	data_out:	out reg8);
end SendReceive;

-- implementao do SendReceive
architecture SendReceive of SendReceive is

type stateSend is (SS0,SS1,SS2,SS3,SS4,SS5,SS6,SS7,SS8);
signal SS, PSS: stateSend;

type stateReceive is (SR0,SR1,SR2,SR3,SR4,SR5,SR6,SR7,SR8,SR9,SR10,SR11);
signal SR, PSR: stateReceive;

signal buf: buff;
signal first,last: reg3;
signal tem_espaco_na_fila : std_logic;

signal countReceive,countSend: reg8;
signal counterWordSend, counterWordReceive: reg3;
signal dataS,dataR: packet;
signal dataR_idx,buffer_idx:integer;
signal dataHigh,dataLow:reg8;

signal stateSR : reg8;

begin
	state <= counterWordReceive & first & buf(7) & buf(6) & buf(5) & buf(4) & buf(3) & buf(2) & buf(1) & dataLow & stateSR & dataR(3) & dataR(2) & dataR(1) & dataR(0);
	
	-- indice do buffer de entrada
	buffer_idx <= CONV_INTEGER(first);
	-- indice do dataR
	dataR_idx <= CONV_INTEGER(counterWordReceive);

	process(reset,clock)
	begin
		if reset='1' then
			SS<=SS0;
			SR<=SR0;
		elsif clock'event and clock='1' then
			SS<=PSS;
			SR<=PSR;
		end if;
	end process;

	tem_espaco_na_fila <= '1' when not((first=x"0" and last=x"7") or (first=last+1)) else '0';
	
	ack_rx <= '1' when  (tem_espaco_na_fila='1' and rx='1') else '0';
	                  
	process(reset, clock)
	begin
		if reset='1' then
			last <= (others=>'0');  
		elsif clock'event and clock='1' then
			if tem_espaco_na_fila='1' and rx='1' then
				buf(CONV_INTEGER(last)) <= data_in;
				last <= last + 1;
			end if;
		end if;
	end process;

-----------------------------------------------------------------------------------------------------	
-- SEND
------------------------------------------------------------------------------------------------------	
	------------------------------------------------------------------------------------------------------	
	-- PARTE COMBINACIONAL PARA DEFINIR O PRXIMO ESTADO DA MQUINA DO SEND WRAPPER
	------------------------------------------------------------------------------------------------------	
	process(SS,avSend,ack_tx,countSend)
	begin
		case SS is
			when SS0 => if avSend ='1' then PSS <= SS1; else PSS <= SS0; end if;
			when SS1 => if ack_tx='1' then PSS <= SS2; else PSS <= SS1; end if;
			when SS2 => PSS <= SS3;
			when SS3 => if ack_tx='1' then PSS <= SS4; else PSS <= SS3; end if;
			when SS4 => PSS <= SS5;
			when SS5 => if ack_tx='1' then PSS <= SS6; else PSS <= SS5; end if;
			when SS6 => if countSend=x"0" then PSS <= SS0; else PSS <= SS7; end if;
			when SS7 => if ack_tx='1' then PSS <= SS8; else PSS <= SS7; end if;
			when SS8 => PSS <= SS5;
		end case;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- executa as aes correspondente ao estado atual da mquina de estados SEND WRAPPER
	------------------------------------------------------------------------------------------------------	
	process (clock)
	begin
		if clock'event and clock='0' then
			case SS is
				when SS0 =>
					busySend <= '0';
					countSend <= (others => '0');
					counterWordSend <= (others => '0');
					dataS <= (others => (others => '0'));
				when SS1 =>
					busySend <= '1';
					dataS <= dataSend;
					data_out <= dataSend(0)(15 downto 8);
					tx<='1';
					countSend<= x"0" & CONV_HEX(CONV_INTEGER(nwordSend)*2-1);
				when SS2 =>
					tx <= '0';
				when SS3 =>
					data_out <= countSend;
					tx<='1';
				when SS4 =>
					tx <= '0';
				when SS5 =>
					data_out <= dataS(CONV_INTEGER(counterWordSend))(7 downto 0);
					tx<='1';
				when SS6 =>
					countSend <= countSend - '1';
					counterWordSend <= counterWordSend + '1';
					tx <='0';
				when SS7 =>
					data_out <= dataS(CONV_INTEGER(counterWordSend))(15 downto 8);
					tx<='1';
				when SS8 =>
					countSend <= countSend - '1';
					tx <='0';
			end case;
		end if;
	end process;

-----------------------------------------------------------------------------------------------------	
-- RECEIVE
------------------------------------------------------------------------------------------------------	
	------------------------------------------------------------------------------------------------------	
	-- PARTE COMBINACIONAL PARA DEFINIR O PRXIMO ESTADO DA MQUINA
	------------------------------------------------------------------------------------------------------	
	process(SR,busyWrapper,first,last,countReceive)
	begin
		case SR is
			when SR0 => PSR <= SR1;
			when SR1 => if first/=last then PSR <= SR2; else PSR <= SR1; end if;
			when SR2 => PSR <= SR3;
			when SR3 => if first/=last then PSR <= SR4; else PSR <= SR3; end if;
			when SR4 => PSR <= SR5;
			when SR5 => if first/=last then PSR <= SR6; else PSR <= SR5; end if;
			when SR6 => PSR <= SR7;
			when SR7 => if countReceive = x"0" and busyWrapper='0' then PSR <= SR11;
				elsif countReceive = x"0" then PSR <= SR10;
				elsif first/=last then PSR <= SR9; 
				else PSR <= SR8; end if;
			when SR8 => if first/=last then PSR <= SR9; else PSR <= SR8; end if;
			when SR9 => PSR <= SR5;
			when SR10 =>  if busyWrapper='0' then PSR <= SR11; else PSR <= SR10; end if;
			when SR11 => PSR <= SR1;
		end case;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- executa as aes correspondente ao estado atual da mquina de estados
	------------------------------------------------------------------------------------------------------	
	process(clock)
	begin
		if clock'event and clock='0' then
			case SR is
				when SR0 =>					
					counterWordReceive <= (others=>'0');
					first <= (others=>'0');
					avReceive <= '0';
					stateSR <= x"00";
				when SR1 =>
					avReceive <= '0';
					counterWordReceive <= (others=>'0');
					stateSR <= x"01";
				when SR2 =>
					dataHigh <= buf(buffer_idx);
					first <= first + '1';
					stateSR <= x"02";
				when SR3 =>
					stateSR <= x"03";
				when SR4 =>
					countReceive <= buf(buffer_idx);
					first <= first + '1';
					stateSR <= x"04";
				when SR5 =>
					stateSR <= x"05";
				when SR6 =>
					dataLow <= buf(buffer_idx);
					countReceive <= countReceive - '1';
					first <= first + '1';
					stateSR <= x"06";
				when SR7 =>
					dataR(dataR_idx) <= dataHigh & dataLow;
					counterWordReceive <= counterWordReceive + '1';
					stateSR <= x"07";
				when SR8 =>
					stateSR <= x"08";
				when SR9 =>
					dataHigh <= buf(buffer_idx);
					countReceive <= countReceive - '1';
					first <= first + '1';
					stateSR <= x"09";
				when SR10 =>
					stateSR <= x"0A";
				when SR11 =>
					avReceive <= '1';
					dataReceive <= dataR;
					stateSR <= x"0B";
				when others=>null;
			end case;
		end if;
	end process;

end SendReceive;