library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use work.packet.all;

-- Retira daqui
--library UNISIM;
--use UNISIM.vcomponents.all;
--use IEEE.VITAL_Timing.all;
-- at aqui para prototipar

-- interface do WrapperBlockram
entity WrapperBlockram is
	port(
	-- debug
	debug:			out std_logic_vector(141 downto 0);

	clock:			in std_logic;
	reset:			in std_logic;
	addressC:		in reg8;
	rx:			in std_logic;
	ack_rx:			out std_logic;
	data_in:		in reg8;
	tx:			out std_logic;
	ack_tx:			in std_logic;
	data_out:		out reg8);
end WrapperBlockram;

-- implementao do WrapperBlockram
architecture WrapperBlockram of WrapperBlockram is

-- interface do Send
component SendReceive is
port(
	state:		out std_logic_vector(141 downto 0);
	clock: 		in std_logic;
	reset: 		in std_logic;
	addressC:	in std_logic_vector(7 downto 0);
	avSend:		in std_logic;
	nwordSend:	in std_logic_vector(2 downto 0);
	busySend:	out std_logic;
	busyWrapper:	in std_logic;
	dataSend:	in packet;
	dataReceive:	out packet;
	avReceive:	out std_logic;
	rx:		in std_logic;
	ack_rx:		out std_logic;
	data_in:	in reg8;
	tx:		out std_logic;
	ack_tx:		in std_logic;
	data_out:	out reg8);
end component;

-- interface do bloco de RAM
--componente da Virtex I
--component RAMB4_S4
--	port(
--	WE, EN, RST, CLK:	in std_logic;
--	ADDR:			in std_logic_vector(9 downto 0);
--	DI:			in std_logic_vector(3 downto 0);
--	DO:			out std_logic_vector(3 downto 0));
--end component;

--componente da Virtex II
component RAMB16_S9 
port	(DI     : in STD_LOGIC_VECTOR (7 downto 0);
        DIP    : in STD_LOGIC_VECTOR (0 downto 0);
        EN     : in STD_ULOGIC;
        WE     : in STD_ULOGIC;
        SSR    : in STD_ULOGIC;
        CLK    : in STD_ULOGIC;
        ADDR   : in STD_LOGIC_VECTOR (10 downto 0);
        DO     : out STD_LOGIC_VECTOR (7 downto 0);
        DOP    : out STD_LOGIC_VECTOR (0 downto 0)); 
end component;

-- interface do processador
component processador is
port(
	ck,rst:  in std_logic;
	dataIN:  in  reg16;
	dataOUT: out reg16;
	address: out reg16;
	ce,rw:   out std_logic;
	waitR8:  in std_logic;
	haltR8:  out std_logic);
end component;

type state is (S0,S1,S2,S3,S4,S5,S6,S7,S8);
signal ES, PES: state;

type stateProc is (SP0,SP1,SP2,SP3,SP4,SP5,SP7,SP8,SP10,SP11,SP12,SP13,SP14,SP15,SP16,SP17,SP18,SP19,SP20,SP21,SP22);
signal SP, PSP: stateProc;

type stateNotify is (SN0,SN1,SN2,SN3);
signal SN,PSN: stateNotify;

type stateMem is (SM0,SM1,SM2,SM3,SM4,SM5,SM6,SM7,SM8);
signal SM, PSM: stateMem;

-- Processador
signal abilityProc,receiveData,receiveNotify: std_logic;
signal ceR8,waitR8, haltR8:	std_logic;
signal addressP,data_inP,data_outP:	reg16;
signal addressMemDist: reg16;
signal timeout, nrepeat: reg4;
signal nciclos,nciclosExecution: reg32;
signal waitfor: reg8;
signal notify: std_logic_vector(10 downto 0);

-- Memoria
signal readMem,writeMem: std_logic;
signal accessBR: std_logic_vector(1 downto 0);
signal Nrw,en,we,Nclock: std_logic;
signal addressCReceive,addressCSource: std_logic_vector(7 downto 0);
signal command: std_logic_vector(15 downto 0);
signal din,dout,data_inBR,data_outBR: std_logic_vector(15 downto 0);
-- VIRTEX
--signal addressBR,addr: reg10;
-- VIRTEX II
signal addressBR,addr: reg11;
signal paridade: std_logic_vector(0 downto 0);
signal ssr: std_ulogic;

-- SendReceive
signal avSend, avReceive, busySend, busyWrapper: std_logic;
signal nwordSend: std_logic_vector(2 downto 0);
signal dataSend, dataReceive: packet;

-- sinais do barramento processador memoria
signal data_inMem,data_outMem: reg16;
signal ce,rw,ack: std_logic;
-- VIRTEX
--signal addressMem: reg10;
-- VIRTEX II
signal addressMem: reg11;


signal 	stateWB: reg8;
signal 	stateSR: std_logic_vector(141 downto 0);

signal debug1: std_logic_vector(100 downto 0);


signal addressSerial : reg8; --provisorio;
begin
	addressSerial <= x"00";

	-- debug
	debug1 <= (others=>'0');
	debug <= accessBR & we & rw & ce & en & stateWB & addr & dout & debug1;

	Nclock <= clock;
	Nrw <= not rw;

	busyWrapper<='0'; --temporariamente

	-- sinais da memria da Virtex II
	paridade <= "1";
	ssr <= '1';



	SR : SendReceive
	port map(
	state => 	stateSR,

	clock => clock,
	reset => reset,
	addressC =>	addressC,
	avSend => avSend,
	nwordSend => nwordSend,
	busySend =>	busySend,
	busyWrapper => busyWrapper,
	dataSend =>	dataSend,
	dataReceive => dataReceive,
	avReceive => avReceive,
	rx => rx,
	ack_rx => ack_rx,
	data_in => data_in,
	tx => tx,
	ack_tx => ack_tx,
	data_out =>	data_out);
	
	-- instanticiando blockram da Virtex I
	--RAM0: RAMB4_S4
	--port map(WE=>we, EN=>en, RST=>reset, CLK=>Nclock, ADDR=>addr, DI=>din(3 downto 0), DO=>dout(3 downto 0));
	--RAM1: RAMB4_S4
	--port map(WE=>we, EN=>en, RST=>reset, CLK=>Nclock, ADDR=>addr, DI=>din(7 downto 4), DO=>dout(7 downto 4));
	--RAM2: RAMB4_S4
	--port map(WE=>we, EN=>en, RST=>reset, CLK=>Nclock, ADDR=>addr, DI=>din(11 downto 8), DO=>dout(11 downto 8));
	--RAM3: RAMB4_S4
	--port map(WE=>we, EN=>en, RST=>reset, CLK=>Nclock, ADDR=>addr, DI=>din(15 downto 12), DO=>dout(15 downto 12));
	
	-- instanticiando blockram da Virtex II
	RAM0: RAMB16_S9
	port map(WE=>we, EN=>en, DIP=>paridade, SSR=>reset, CLK=>Nclock, ADDR=>addr, DI=>din(7 downto 0), DO=>dout(7 downto 0));
	RAM1: RAMB16_S9
	port map(WE=>we, EN=>en, DIP=>paridade, SSR=>reset, CLK=>Nclock, ADDR=>addr, DI=>din(15 downto 8), DO=>dout(15 downto 8));

	proc : processador
	port map(
	ck => clock,
	rst => reset,
	dataIN => data_inP,
	dataOUT => data_outP,
	address => addressP,
	ce => ceR8,
	rw => rw,
	waitR8 => waitR8,
	haltR8 => haltR8);

	process(reset,clock)
	begin
		if reset='1' then
			ES<=S0;
			SM<=SM0;
			SP<=SP0;
			SN<=SN0;
		elsif clock'event and clock='1' then
			ES<=PES;
			SM<=PSM;
			SP<=PSP;
			SN<=PSN;
		end if;
	end process;

----------------------------------------------------------------------------------------
-- Maquina de estado que identifica se o pacote  para o processador ou para a memoria.
-- A identificao  realizada atravs do comando do pacote.
--           Comando 00 (leitura) - pacote para a memoria
--           Comando 01 (escrita) - pacote para a memoria
--           Comando 02 (habilitacao) - pacote para o processador
--           Comando 08 (notify) - pacote para o processador
--           Comando 09 (retorno da leitura) - pacote para o processador
-- Quando o pacote  para a memria a mquina de estados SM  ativa pelo sinal readMem
-- ou writeMem.
----------------------------------------------------------------------------------------
	process(ES,reset,avReceive,dataReceive,SM,accessBR)
	begin
		case ES is
			when S0 => PES<= S1;
			when S1 => if avReceive='1' then
					if (dataReceive(1)=x"01" or dataReceive(1)=x"00") and SM=SM0 then PES<=S2;
					elsif dataReceive(1)=x"02" then PES <= S6;
					elsif dataReceive(1)=x"09" or dataReceive(1)=x"04" then PES <= S7;
					elsif dataReceive(1)=x"08" then PES <= S8;
					else PES <= S1;
					end if;
				end if;
			when S2 => if accessBR=x"00" then
						if dataReceive(1)=x"01" then PES<=S4; else PES<=S5; end if;
					   else PES<=S3; end if;
			when S3 => if accessBR=x"00" then -- repete porque nao pode ficar no estado 1.
						if dataReceive(1)=x"01" then PES<=S4; else PES<=S5; end if;
					   else PES<=S3; end if;
			when others => PES<= S1;
		end case;
	end process;

	process (clock)
	begin
		if clock'event and clock='0' then
			case ES is
				when S0 =>
					writeMem <= '0';
					readMem <= '0';
					abilityProc <= '0';
					receiveData <= '0';
					receiveNotify <= '0';
					addressCSource <= (others => '0');
					command <= (others => '0');
					addressBR <= (others => '0');
					data_inBR <= (others => '0');
				when S1 =>
					writeMem <= '0';
					readMem <= '0';
					abilityProc <= '0';
					receiveData <= '0';
					receiveNotify <= '0';
				when S2 =>
					-- "primeiro dado" [addressCReceive][addressCSource]
					addressCSource<= dataReceive(0)(7 downto 0);
					-- "segundo dado" [command]
					command <= dataReceive(1);
					-- "terceiro dado" [000000][addressBR]
					-- VIRTEX
					--addressBR <= dataReceive(2)(9 downto 0);
					-- VIRTEX II
					addressBR <= dataReceive(2)(10 downto 0);
					-- "quarto dado" [dataBR]
					data_inBR <= dataReceive(3);
				when S4 => writeMem <= '1';
				when S5 => readMem <= '1';
				when S6 => abilityProc <= '1';
				when S7 => receiveData <= '1';
				when S8 => receiveNotify <= '1';
				when others => null;
			end case;
		end if;
	end process;


----------------------------------------------------------------------------------------
-- MEMRIA
----------------------------------------------------------------------------------------
	----------------------------------------------------------------------------------------
	-- Maquina de estado que executa comandos destinados a Memria.
	----------------------------------------------------------------------------------------
	process(SM,writeMem,readMem,busySend)
	begin
		case SM is
			when SM0 => if writeMem='1' then PSM <= SM1;
						elsif readMem='1' then PSM <= SM2;
						else PSM <= SM0; end if;
			when SM1 => PSM<=SM0; -- escrevendo na memoria
			when SM2 => PSM<=SM3; -- lendo da memoria
			when SM3 => PSM<=SM4; -- recebendo dado lido
			when SM4 => if busySend='0' then PSM <= SM5; else PSM<= SM4; end if; --esperando liberacao do Send			
			when SM5 => PSM<=SM0; --enviando pacote			
			when others => PSM<=SM0;
		end case;
	end process;
	

	----------------------------------------------------------------------------------------					
	-- ACCESS MEMORY	
	---------------------------------------------------------------------------------------	
	process(clock,reset,accessBR)
	begin
		if reset='1' then ack<='0';
		elsif clock'event and clock='0' then
			if accessBR="01" then ack<='1';	else ack<='0'; end if;
		end if;
	end process;

	process(clock,reset,SM)
	begin
		if reset='1' then
			data_outBR<=(others=>'0');
		elsif clock'event and clock='0' then
			if SM=SM3 then data_outBR<= dout; end if;
		end if;
	end process;


	stateWB<= x"01" when SM=SM1 else
			x"02" when SM=SM2 else
			x"03" when SM=SM3 else
			x"04" when SM=SM4 else
			x"05" when SM=SM5 else
			x"00";


	accessBR <= "10" when SM=SM1 or SM=SM2 else --  SM2 lendo da blockram e SM1 escrevendo na blockram.
				"01" when ce='1' else
		    	"00"; 

	we <= Nrw when accessBR="01" else
	      '1' when SM=SM1 else
	      '0';

	en <= '1' when accessBR/="00" else '0';

	addr <= addressMem when accessBR="01" else
    	    addressBR when accessBR="10" else
        	(others=>'0');

	din <= data_inMem when accessBR="01" else
    	   data_inBR when accessBR="10" else
		   (others=>'0');

	data_outMem<= dout when accessBR="01" else (others=>'0');

------------------------------------------------------------------------------------------------------	
-- Barramento processador Memoria
------------------------------------------------------------------------------------------------------	
	-- VIRTEX
	--addressMem <= addressP(9 downto 0);
	-- VIRTEX II
	addressMem <= addressP(10 downto 0);
	
	data_inMem <= data_outP;
  
	ce <= '1' when ceR8='1' and addressP(15 downto 10)="000000" else
		  '0';

	data_inP <= dataReceive(2) when SP=SP11 else -- retorno do scanf ou retorno de leitura a memoria distribuida
			data_outMem;


	addressMemDist <= addressP - x"0400" when (addressP>x"03FF" and addressP<x"0800" and (SP=SP4 or SP=SP8)) else
				  addressP - x"0800" when (addressP>x"07FF" and addressP<x"0C00" and (SP=SP4 or SP=SP8)) else
				  (others=>'0');
	
	addressCReceive <= ADDRESSN10 when (addressP>x"03FF" and addressP<x"0800" and (SP=SP4 or SP=SP8)) else
			   ADDRESSN01 when (addressP>x"07FF" and addressP<x"0C00" and (SP=SP4 or SP=SP8)) else
			   (others=>'0');


------------------------------------------------------------------------------------------------------	
-- PROCESSADOR
------------------------------------------------------------------------------------------------------	

	process(clock,SP,ceR8,ack,waitR8)
	begin
		if SP=SP0 then
			waitR8<='1';
		elsif clock'event and clock='0' then
			if (ceR8 ='1' and waitR8='0') or SP=SP12 or SP=SP15 then
				waitR8<= '1';
			elsif (ack='1' or SP=SP1 or SP=SP11) then
				waitR8<= '0';
			end if;
		end if;
	end process;

	process(reset,clock,SP)
	begin
		if reset='1' then
			nciclos <= (others=>'0');
		elsif clock'event and clock='1' then
			if SP/=SP0 then
				nciclos <= nciclos + '1';
			end if;
		end if;
	end process;

	------------------------------------------------------------------------------------------------------	
	-- PARTE COMBINACIONAL PARA DEFINIR O PRXIMO ESTADO DA MQUINA QUE RECEBE OS NOTIFYs.
	------------------------------------------------------------------------------------------------------	
	process(SN,SP,receiveNotify)
	begin
		case SN is
			-- reseta o notify
			when SN0 => PSN <= SN1; 
			-- chegada do comando Notify
			when SN1 => if receiveNotify='1' then PSN <= SN2; 
					elsif SP=SP14 then PSN <= SN3;
					else PSN <= SN1; end if;
			-- coloca notify ativo na posicao do core que envio o notify
			when SN2 => PSN <= SN1;
			-- coloca notify desativo na posicao do waitfor 
			when SN3 => PSN <= SN1;
		end case;
	end process;
	
	------------------------------------------------------------------------------------------------------	
	-- executa as aes correspondente ao estado atual da mquina de estados do notify
	------------------------------------------------------------------------------------------------------	
	process (clock)
	begin
		if clock'event and clock='0' then
			case SN is
				when SN0 =>
					notify<=(others=>'0');
				when SN2 =>
					notify(CONV_INDEXNOTIFY(dataReceive(0)(7 downto 0)))<='1';
				when SN3 =>
					notify(CONV_INDEXNOTIFY(waitfor))<='0'; 
				when others=> null;
			end case;
		end if;				
	end process;
	

	----------------------------------------------------------------------------------------
	-- Maquina de estado que executa comandos destinados ao processador.
	----------------------------------------------------------------------------------------
	process(SP,abilityProc,receiveData,haltR8,ceR8,rw,addressP,busySend,waitfor,notify)
	begin
		case SP is
			-- habilitacao do processador
			when SP0 => if abilityProc='1' then PSP <= SP1; else PSP <= SP0; end if;
			-- processando
			when SP1 => PSP <= SP2;
			when SP2 => if haltR8='1' then PSP <= SP15;
						elsif ceR8='1' and (addressP = x"FFFE") then PSP <= SP12;								       -- execution time
						elsif ceR8='1' and (addressP = x"FFFF") and rw='0' and busySend='0' then PSP <= SP3;	   -- saida
						elsif ceR8='1' and (addressP = x"FFFD") and busySend='0' then PSP <= SP5;					   -- envia notify
						elsif ceR8='1' and (addressP = x"FFFF") and rw='1' and busySend='0' then PSP <= SP7;	   -- entrada
						elsif ceR8='1' and (addressP > x"03FF") and rw='1' and busySend='0' then PSP <= SP8;	   -- leitura
						elsif ceR8='1' and (addressP > x"03FF") and rw='0' and busySend='0' then PSP <= SP4;	   -- escrita
						--caso busySend seja '1'
						elsif ceR8='1' and (addressP = x"FFFF") and rw='0' then PSP <= SP18;	   -- saida
						elsif ceR8='1' and (addressP = x"FFFD") then PSP <= SP20;					   -- envia notify
						elsif ceR8='1' and (addressP = x"FFFF") and rw='1' then PSP <= SP21;	   -- entrada
						elsif ceR8='1' and (addressP > x"03FF") and rw='1' then PSP <= SP22;	   -- leitura
						elsif ceR8='1' and (addressP > x"03FF") and rw='0' then PSP <= SP19;	   -- escrita

						else PSP<= SP2;
						end if;
    		-- armazena dados do printf
			when SP3 => PSP <= SP1;
			-- armazena dados para escrever em memoria distribuida
			when SP4 => PSP <= SP1;
			-- NOTIFY - monta o pacote do notify
			when SP5 => PSP <= SP1;
			-- armazena dados do scanf 
			when SP7 => PSP <= SP10;
			-- armazena dados para ler da memoria distribuida
			when SP8 => PSP <= SP10;
			-- esperando dado
			when SP10 => if receiveData='1' then PSP <= SP11; else PSP <= SP10; end if;
			-- recebendo dado
			when SP11 => PSP <= SP1;
			-- WAIT - armazena o addressC de que se espera o NOTIFY
			when SP12 => PSP<= SP13; 
			-- WAIT - verifica se chegou o notify que se esta esperando
			when SP13 => if notify(CONV_INDEXNOTIFY(waitfor))='1' then PSP<=SP14; else PSP<=SP13; end if; 	
			-- WAIT - chegou o notify
			when SP14 => PSP<=SP1;
			-- captura o numero de ciclos de clock da execucao do processador
			when SP15 => if busySend='0' then PSP <= SP17; else PSP<= SP16; end if;
			-- armazena dados do comando execution time
			when SP16 => if busySend='0' then PSP <= SP17; else PSP<= SP16; end if;
			-- ativando o send e finalizando o processamento
			when SP17 => PSP <= SP0;
			--estados de espera do busySend
			when SP18 => if busySend='0' then PSP <= SP3; else PSP<= SP18; end if;
			when SP19 => if busySend='0' then PSP <= SP4; else PSP<= SP19; end if;
			when SP20 => if busySend='0' then PSP <= SP5; else PSP<= SP20; end if;
			when SP21 => if busySend='0' then PSP <= SP7; else PSP<= SP21; end if;
			when SP22 => if busySend='0' then PSP <= SP8; else PSP<= SP22; end if;
			when others => PSP <= SP1;
		end case;
	end process;
	
	------------------------------------------------------------------------------------------------------	
	-- executa as aes correspondente ao estado atual da mquina de estados
	------------------------------------------------------------------------------------------------------	
	process (clock)
	begin
		if clock'event and clock='0' then
			case SP is
				when SP0 =>
					waitfor<=(others=>'0');
				-- WAIT - armazenado o addressC de que se espera o NOTIFY
				when SP12 =>
					waitfor <= data_outP(7 downto 0);
				-- armazena o numero de ciclos de clock a execucao gastou
				when SP15 =>
					nciclosExecution<= nciclos;
				when others=> null;
			end case;
		end if;
	end process;

------------------------------------------------------------------------------------------------------	
-- controle do sinais do envio de pacote para o SENDRECEIVE (avSend,nwordSend,dataSend)
------------------------------------------------------------------------------------------------------	
	
	process(clock,reset)
	begin
		if reset='1' then
			avSend <= '0';
			nwordSend <= (others=>'0');
			dataSend(0) <= (others=>'0');
			dataSend(1) <= (others=>'0');
			dataSend(2) <= (others=>'0');
			dataSend(3) <= (others=>'0');
		elsif clock'event and clock='0' then
			if SM=SM5 then -- retorno da leitura na memoria distribuida
				-- habilita o envio de pacote
				avSend <= '1';
				nwordSend <= "011";
				dataSend(0) <= addressCSource & addressC;
				dataSend(1) <= x"0009";
				dataSend(2) <= data_outBR;
			elsif SP=SP3 then -- printf
				-- habilita o envio de pacote
				avSend <= '1';
				-- numero de palavras do printf
				nwordSend <= "011";
				-- "primeiro dado" [core destino][core origem]
				dataSend(0) <= addressSerial & addressC;
				-- "segundo dado" [command]
				dataSend(1) <= x"0003";
				-- "terceiro dado" [data_outP]
				dataSend(2) <= data_outP;
			elsif SP=SP4 then -- escrita em memoria distribuida
				-- habilita o envio de pacote
				avSend <= '1';
				-- numero de palavras do acesso
				nwordSend <= "100";
				-- "primeiro dado" [core destino][core origem]
				dataSend(0) <= addressCReceive & addressC;
				-- "segundo dado" [command]
				dataSend(1) <= x"0001";
				-- "terceiro dado" [address]
				dataSend(2) <= addressMemDist;
				-- "quarto dado" [data]
				dataSend(3) <= data_outP;	  
			elsif SP=SP5 then -- notify
				-- habilita o envio de pacote
				avSend <= '1';
				-- numero de palavras do notify
				nwordSend <= "010";
				-- "primeiro dado" [core destino][core origem]
				dataSend(0) <= data_outP(7 downto 0) & addressC;
				-- "segundo dado" [command]
				dataSend(1) <= x"0008";
			elsif SP=SP7 then -- scanf
				-- habilita o envio de pacote
				avSend <= '1';
				-- numero de palavras do scanf
				nwordSend <= "010";
				-- "primeiro dado" [core destino][core origem]
				dataSend(0) <= addressSerial & addressC;
				-- "segundo dado" [command]
				dataSend(1) <= x"0004";
			elsif SP=SP8 then -- leitura da memoria distribuida
				-- habilita o envio de pacote
				avSend <= '1';
				-- numero de palavras da leitura distribuida
				nwordSend <= "011";
				-- "primeiro dado" [core destino][core origem]
				dataSend(0) <= addressCReceive & addressC;
				-- "segundo dado" [command]
				dataSend(1) <= x"0000";
				-- "terceiro dado" [address]
				dataSend(2) <= addressMemDist;
			elsif SP=SP17 then -- execution time
				-- habilita o envio de pacote
				avSend <= '1';
				-- numero de palavras do execution time
				nwordSend <= "100";
				-- "primeiro dado" [core destino][core origem]
				dataSend(0) <= addressSerial & addressC;
				-- "segundo dado" [command]
				dataSend(1) <= x"0006";
				-- "terceiro dado" [nciclosExecution(31 downto 16)]
				dataSend(2) <= nciclosExecution(31 downto 16);
				-- "quarto dado" [nciclosExecution(31 downto 16)]
				dataSend(3) <= nciclosExecution(15 downto 0);
			else
				-- desabilita o envio de pacote
				avSend <= '0';
			end if;
		end if;
	end process;

end WrapperBlockram;